from . import colors
from .application import Application
from .key import Key
from .panel import Panel
from .scrollpanel import ScrollPanel
from .inputline import InputLine, Completion
from .inputline import SingleWordCompletion, SingleWordCompletionWrapper
from .shellcontext import ShellContext
from .grid import Grid
from .progress import ProgressBar



#!/usr/bin/env python3
import random
import curses
from cursedspace import Application, Key, Panel, ProgressBar


class DemoApplication(Application):
    def __init__(self):
        super().__init__()

    def main(self):
        # Lets make a selfupdating application by setting a timeout for key-capture
        self.screen.timeout(400)
        
        self.init_bars()
        while True:
            self.draw()

            key = self.read_key()

            if key == Key.RESIZE:
                self.resize()
            elif key in [Key.ESCAPE, "q", "^C"]:
                break

    def draw(self):
        # Do update including description for first bar and just update the rest
        for bar in self.bars[1:]:
            bar.update((bar.progress + 1) % 100)
        val = (self.bars[0].progress + 1) % 100
        self.bars[0].update(val, f'Bar 0: {val:03d}')

        for bar in self.bars:
            bar.paint()
            # try:
            #     bar.paint()
            # except:
            #     pass
        curses.doupdate()

    def init_bars(self):
        # Some border combinations to try out!
        border_list = [
            Panel.BORDER_ALL,
            Panel.BORDER_TOP,
            Panel.BORDER_BOTTOM,
            Panel.BORDER_LEFT,
            Panel.BORDER_RIGHT,
            Panel.BORDER_TOP + Panel.BORDER_LEFT,
            Panel.BORDER_TOP + Panel.BORDER_RIGHT,
            Panel.BORDER_BOTTOM + Panel.BORDER_LEFT,
            Panel.BORDER_BOTTOM + Panel.BORDER_RIGHT,
            Panel.BORDER_BOTTOM + Panel.BORDER_TOP,
            Panel.BORDER_LEFT + Panel.BORDER_RIGHT,
        ]

        # Initiate the bars
        h, w = self.size()
        y = 1
        self.bars = []
        for i in range(h-2):
            bar = ProgressBar(
                self,
                border = border_list[i] if i < len(border_list) else Panel.BORDER_NONE,
                width = w - 2,
                pos = (y, 1),
                description = f'Bar {i}',
                color = random.randint(1, 8),
            )
            # Place next bar below this one, taking border size into account
            y += bar.height
            self.bars.append(bar)

            # Dont place too many
            if y > h - 5:
                break

        # Give random initial value
        for bar in self.bars:
            bar.update(random.randint(0, 100))

    def resize(self):
        h, w = self.size()
        for bar in self.bars:
            bar.resize(h, w)


# run the application
DemoApplication().run()



import collections
import curses
import logging


ColorPair = collections.namedtuple("ColorPair", ['foreground', 'background'], defaults=[-1])


# predefined color identifiers
DEFAULT = ColorPair(-1)
RED = ColorPair(curses.COLOR_RED)
GREEN = ColorPair(curses.COLOR_GREEN)
YELLOW = ColorPair(curses.COLOR_YELLOW)
BLUE = ColorPair(curses.COLOR_BLUE)
MAGENTA = ColorPair(curses.COLOR_MAGENTA)
CYAN = ColorPair(curses.COLOR_CYAN)
BLACK = ColorPair(curses.COLOR_BLACK)
WHITE = ColorPair(curses.COLOR_WHITE)


# internal color numbers, ColorPair -> (internal_color_number, curses.color_pair(internal_color_number))
_color_pairs = {}


def as_colorpair(value):
    """Try to turn value into a ColorPair"""
    if isinstance(value, ColorPair):
        return value

    if isinstance(value, int):
        return ColorPair(value)
    elif isinstance(value, tuple) and len(value) == 2 and all([isinstance(v, int) for v in value]):
        return ColorPair(*value)
    elif isinstance(value, tuple) and all([v is None for v in value]):
        return ColorPair(-1, -1)
    
    raise TypeError(f"Expected int or (int, int), but got {type(value)}")


def register(colorpair):
    """Register a color pair to be used in curses later

    Returns True upon success, False upon failure (logs to 'warning')"""
    global _color_pairs
    colorpair = as_colorpair(colorpair)
    next_id = max([0] + [v[0] for v in _color_pairs.values()]) + 1
    try:
        curses.init_pair(next_id, *colorpair)
        _color_pairs[colorpair] = (next_id, curses.color_pair(next_id))
    except Exception as exc:
        logging.warning(f"Failed to register color {colorpair}: {exc}")
        return False
    return True


def attr(color):
    """Return a curses-like attribute value for 'color'

    'color' may be one of the Colors constants or a tuple
    (foreground, background) of Colors constants.
    """
    global _color_pairs
    color = as_colorpair(color)

    if color == (-1, -1):
        return 0

    if color not in _color_pairs:
        if not register(color):
            # fallback to default colors
            return curses.color_pair(0)
    return _color_pairs[color][1]


def init_colors():
    """Initializes the basic foreground colors and sets up color pairs for each"""
    curses.use_default_colors()
    register(DEFAULT)
    register(BLACK)
    register(WHITE)
    register(RED)
    register(GREEN)
    register(YELLOW)
    register(BLUE)
    register(MAGENTA)
    register(CYAN)




from .panel import Panel


class Grid:
    """A generic grid of panel-type objects that allows for any rectangular tiling.
    """

    def __init__(self, app, rows, cols, size=None, column_sep=0, row_sep=0):
        self.app = app
        self.cols = cols
        self.rows = rows
        self.size = size
        self.panels = []
        self._panel_data = []
        self.keys = []
        self.column_sep = column_sep
        self.row_sep = row_sep

    def fill(self, **kwargs):
        """Fill the Grid with panels"""
        for i in range(self.cols):
            for j in range(self.rows):
                self.add_panel(j, i, **kwargs)

    def cell_size(self):
        height, width = self.size
        return height//self.rows - self.row_sep*self.rows, width//self.cols - self.column_sep*self.cols

    def add_panel(self, rows, cols, key=None, panel_class=Panel, args=None, kwargs=None):
        """Add a panel to the grid.

        cols and rows denote the grid-position of the panel and 
        can either be integer indecies or a slice to span 
        multiple grid-cells.

        Returns the panel object that is being stored in the grid.

        Args and kwargs denote the additional arguments to supply 
        to the panel instansiation.
        """
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        if key is not None and key in self.keys:
            raise ValueError(f'Key "{key}" already exists in Grid')

        pos = [None, None]
        size = [None, None]
        if isinstance(rows, slice):
            rows = list(range(self.rows))[rows]
            size[0] = len(rows)
            pos[0] = rows[0]
        else:
            size[0] = 1
            pos[0] = rows

        if isinstance(cols, slice):
            cols = list(range(self.cols))[cols]
            size[1] = len(cols)
            pos[1] = cols[0]
        else:
            size[1] = 1
            pos[1] = cols

        self._panel_data.append([size, pos])
        panel = panel_class(
            self.app, 
            *args, 
            **kwargs
        )

        self.panels.append(panel)
        self.keys.append(key)

    def remove_panel(self, key):
        """Remove a panel from the grid."""
        if isinstance(key, str):
            ind = self.keys.index(key)
        elif isinstance(key, int):
            ind = key
        elif isinstance(key, Panel):
            ind = self.panels.index(key)
        else:
            raise TypeError(f'Key type {type(key)} not recoginzed for deleting panel from Grid')

        del self.panels[ind]
        del self._panel_data[ind]
        del self.keys[ind]

    def resize(self, height, width):
        """Resize the grid"""
        self.size = (height, width)
        cy, cx = self.cell_size()

        for panel, (size, pos) in zip(self.panels, self._panel_data):
            sy = size[0]*cy + (size[0] - 1)*self.row_sep
            sx = size[1]*cx + (size[1] - 1)*self.column_sep
            py = pos[0]*(cy + self.row_sep) + self.row_sep
            px = pos[1]*(cx + self.column_sep) + self.column_sep

            panel.resize(sy, sx)
            panel.move(py, px)

    def refresh(self, force=False):
        """Refresh the grid"""
        for panel in self.panels:
            panel.refresh(force=force)

    def paint(self, clear=False):
        """Paint the grid"""
        for panel in self.panels:
            panel.paint(clear=clear)

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.panels[self.keys.index(key)]
        elif isinstance(key, int):
            return self.panels[key]
        else:
            raise TypeError(f'Key type {type(key)} not recoginzed for Grid indexing')



import logging
import locale
import curses
import sys

from . import colors
from .key import Key


class Application:
    """A generic application with curses

    Use self.screen for drawing operations.
    """
    def __init__(self):
        self.screen = None
        self.is_initialized = False
        self.preferred_encoding = None
        self._cursor_set = False
        self._original_cursor = 1

    def refresh(self, force=False):
        """Refreshes the screen

        If force is set to True, it will touch the screen first
        to enforce a full refresh"""
        if force:
            self.screen.touchwin()
        self.screen.noutrefresh()

    def size(self):
        """Refreshes the size of the screen and returns the (height,width) tuple"""
        curses.update_lines_cols()
        return self.screen.getmaxyx()

    def main(self):
        """Your entry point for the main loop"""
        raise NotImplementedError()

    def read_key(self):
        """Read the next key press from the system

        Returns a cursedspace.Key instance"""
        return Key.read(self.screen)

    def run(self):
        """The main function to call when actually running the application"""
        exception = None

        try:
            self.init_curses()
            self.main()
        except Exception as exc:
            exception = exc

        self.end_curses()

        if exception is not None:
            raise exception

        return 0

    def show_cursor(self, state):
        """Whether or not to show the cursor

        This is a wrapper for curses.curs_set, but will remember the original
        state of the cursor and restore it when the application ends, so you
        don't have to.
        """
        if isinstance(state, bool):
            state = 1 if state else 0

        try:
            previous_state = curses.curs_set(state)
        except curses.error:
            return

        if not self._cursor_set:
            self._original_cursor = previous_state

    def init_curses(self):
        """Will be called prior to using any curses operations in self.run()

        Normally you should not have to call this."""
        if self.is_initialized:
            return

        locale.setlocale(locale.LC_ALL, '')
        self.preferred_encoding = locale.getpreferredencoding()

        self.screen = curses.initscr()

        if curses.has_colors():
            curses.start_color()
            colors.init_colors()

        curses.noecho()
        curses.cbreak()
        self.screen.keypad(True)

        curses.def_prog_mode()

        self.is_initialized = True

    def end_curses(self):
        """Will be called after the end of the application

        This function will restore the terminal and cursor state prior
        to the launch of the application and clear the screen.
        """
        if not self.is_initialized:
            return

        self.screen.clear()
        self.screen.refresh()

        if self._cursor_set:
            try:
                curses.curs_set(self._original_cursor)
            except curses.error:
                pass

        curses.reset_shell_mode()

        try:
            curses.endwin()
        except curses.error as exc:
            logging.info("Ending curses failed: %s", exc)

    def set_term_title(self, title):
        """Attempt to set the title of the terminal emulator"""
        # trying our best
        sys.stdout.write(f'\x1b]0;{title}\x07')
        sys.stdout.write(f'\x1b]2;{title}\x07')
        sys.stdout.flush()

    def __delete__(self):
        self.end_curses()



"""A single line of user input with optional completion"""
import logging
import curses
import string

from . import colors
from .panel import Panel
from .scrollpanel import ScrollPanel
from .key import Key


class Completion:
    """A completion/suggestion box for InputLine

    This allows you to add autocompletion/suggestions to an InputLine
    component.

    You will have to derive your own variant from this though and implement
    `update`.

    Have a look at the sample `SingleWordCompletion` to see what that could
    look like.
    """
    class Suggestion:
        def __init__(self, text, label=None):
            self.text = text
            self.label = label

        def __str__(self):
            if self.label is not None:
                return self.label
            return self.text

        def __lt__(self, other):
            return str(self) < str(other)

    COLOR = colors.DEFAULT

    KEYS_NEXT_ALTERNATIVE = [Key.DOWN, "^N", Key.TAB]
    KEYS_PREVIOUS_ALTERNATIVE = [Key.UP, "^P"]
    KEYS_SELECT_ALTERNATIVE = [Key.RETURN]
    KEYS_CANCEL_SUGGESTIONS = [Key.ESCAPE, "^C"]

    class SuggestionPanel(ScrollPanel):
        def __init__(self, parent):
            super().__init__(parent.app)
            self.color = parent.COLOR

        def do_paint_item(self, y, x, maxwidth, is_selected, item):
            attr = curses.A_REVERSE
            if is_selected:
                attr = curses.A_NORMAL
            attr += colors.attr(self.color)
            if isinstance(item, Completion.Suggestion):
                label = item.label if item.label is not None else item.text
            else:
                label = str(item)
            label = label[:maxwidth]
            label = label + " "*(maxwidth-len(label)+1)
            try:
                self.win.addstr(y, x, label[:maxwidth], attr)
            except curses.error:
                pass

    def __init__(self, inputline):
        self.app = inputline.app
        self.inputline = inputline
        self.suggestion_box_type = self.SuggestionPanel
        # the scrollpanel with the suggestions
        self.suggestions = None

    def handle_key(self, key):
        """When InputLine receives a key, it will call this function

        Be sure to return False if you do not consume the key, because
        only then will InputLine handle the key itself.
        """
        if not self.is_visible:
            return False

        handled = self.suggestions.handle_key(key)

        if not handled:
            if key in self.KEYS_CANCEL_SUGGESTIONS:
                self.close()
                handled = True
            elif key in self.KEYS_SELECT_ALTERNATIVE:
                self.alternative_selected(self.suggestions.selected_item)
                self.close()
                handled = True

            if handled:
                self.app.refresh(True)

        return handled

    def alternative_selected(self, alternative):
        """Called when the user selected an alternative from the list of options"""
        if isinstance(alternative, Completion.Suggestion):
            alternative = alternative.text
        self.inputline.replace_word(alternative)

    def paint(self, clear=False):
        if self.is_visible:
            self.suggestions.paint(True)

    def update(self, y, x):
        """Called by the InputLine when the user changed the text

        ``y`` and ``x`` are the absolute coordinates of the cursor
        in respect to the app.screen window.

        To access the current text and the cursor position in the
        text, refer to ``self.inputline.text`` and ``self.inputline.cursor``.
        Or get the current word through ``self.inputline.current_word``.

        In this function you must compile the possible suggestion alternatives
        and call ``self.set_alternatives(alternatives, y, x)``.
        Afterwards you will have to call ``self.app.paint``, too.

        Or, if no alternatives can be suggested, make sure to call
        ``self.close``.
        """
        raise NotImplementedError()

    def autosize(self, y, x):
        """Auto resize the suggestion box with the anchor being at y,x"""
        if not self.is_visible:
            return

        screen_h, screen_w = self.app.size()

        space_below = screen_h - (y+1)
        space_above = y-1
        space_right = screen_w - (x+1)
        height = min([10, len(self.suggestions.items), max(space_below, space_above)])
        width = max([len(item.label)
                     if isinstance(item, Completion.Suggestion) else len(str(item))
                     for item in self.suggestions.items])

        py = y
        px = x

        if width >= screen_w:
            width = 2*width//3

        if space_right < width:
            if x >= width:
                px = x-width
            else:
                px = screen_w - width

        if space_below < height and space_above < height:
            if space_below > space_above:
                height = space_below
            else:
                height = space_above

        if space_below >= height:
            py = y+1
        elif space_above >= height:
            py = y - height

        self.suggestions.resize(height, width)
        self.suggestions.move(py, px)

    def set_alternatives(self, items, pos=None):
        if not self.is_visible:
            if pos is None:
                pos = self.inputline.absolute_cursor
            self.suggestions = self.suggestion_box_type(self)
            if isinstance(self.suggestions, ScrollPanel):
                self.suggestions.wrapping = True
                self.suggestions.SCROLL_NEXT = self.KEYS_NEXT_ALTERNATIVE
                self.suggestions.SCROLL_PREVIOUS = self.KEYS_PREVIOUS_ALTERNATIVE
                self.suggestions.SCROLL_TO_END = []
                self.suggestions.SCROLL_TO_START = []
                self.suggestions.SCROLL_NEXT_PAGE = []
                self.suggestions.SCROLL_PREVIOUS_PAGE = []
                self.suggestions.SCROLL_MARGIN = 1
        else:
            if pos is None:
                pos = self.suggestions.pos[:]
        self.suggestions.items = items[:]
        self.autosize(*pos)

    @property
    def is_visible(self):
        return self.suggestions is not None

    def close(self):
        """Close the suggestions panel

        Returns True if the suggestion box was destroyed, False if nothing was done"""
        if self.is_visible:
            self.suggestions.destroy()
            self.suggestions = None
            return True
        return False


class SingleWordCompletion(Completion):
    """An example implementation of a completion

    This will provide autocompletion/suggestion for the first word
    that the user types/might type in the input line.

    To use it, pass SingleWordCompletionWrapper(list_of_words) to the input
    line like this:

        words = ["brick", "elbow", "pudding"]
        edit = InputLine(self, self.size()[1], (0, 0),
                         completion=SingleWordCompletionWrapper(words))

    or instead assign the `completion` property afterwards:

        edit = InputLine(self, self.size()[1], (0, 0))
        edit.completion = SingleWordCompletion(edit, words)

    You can also use Completion.Suggestion to select by label and insert
    other text:

        suggestions = [Completion.Suggestion(*pair)
                       for pair in [("Spanish Inquisition", "nobody"),
                                    ("I'm not dead!", "parrot")]]
        edit.completion = SingleWordCompletion(edit, suggestions)
    """
    def __init__(self, inputline, words):
        super().__init__(inputline)
        self.words = words

    def update(self, y, x):
        text = self.inputline.text[:self.inputline.cursor]
        if text.endswith(' '):
            text = ''
        else:
            text = text.split(' ')[-1]

        options = [word
                   for word in sorted(self.words, key=lambda a: str(a))
                   if (isinstance(word, Completion.Suggestion) and word.label.startswith(text))
                      or str(word).startswith(text)]
        if len(options) == 1 and \
           ((isinstance(options[0], Completion.Suggestion) and options[0].label == text)
                   or str(options[0]) == text):
            # only available option typed out
            self.close()
        elif len(options) > 0:
            # there are options? show them
            self.set_alternatives(options, (y, x))
        elif self.is_visible:
            # no options, but the box is still visible? close it
            self.close()


def SingleWordCompletionWrapper(words):
    """Convenience wrapper for SingleWordCompletion"""
    def wrapped(inputline):
        return SingleWordCompletion(inputline, words)
    return wrapped


class InputLine(Panel):
    """A panel that allows for user input as a text input field does.

    Example use case::

        # in your Application's run function:
        edit = InputLine(self, self.size()[1], (0, 0))

        while True:
            key = self.read_key()
            if key == Key.RESIZE:
                edit.resize(self.size()[1])
            elif key == Key.RETURN:
                # user accepted input, check edit.text
                break
            elif key == Key.ESCAPE:
                # user cancelled input
                break
            else:
                edit.handle_key(key)

    The input line allows the use of a completion box. To use it, just provide
    an instance of Completion (you will have to derive your own though):

        class MyFancyCompletion(Completion):
            def update(self, y, x):
                self.set_alternatives(["fancy"], (y, x))

        edit = InputLine(self, self.size()[1], (0, 0), completion=MyFancyWordCompletion)

    Or assign the InputLine.completion parameter later to your instance:

        edit = InputLine(self, self.size()[1], (0, 0))
        edit.completion = MyFancyCompletion(edit)
    """
    def __init__(self, app, width, pos, text="", prefix="", background=' ', completion=None):
        self.completion = None
        super().__init__(app, (1, width), pos)
        self.completion = completion(self) if completion is not None else None
        self.text = text
        self.offset = 0
        self.cursor = 0
        self.margin = 2
        self.background = background
        self.prefix = prefix
        self.read_only = False

    def paint(self, clear=False):
        self.border = Panel.BORDER_NONE
        super().paint(clear)
        self.scroll()
        visible_text = self.text[self.offset:]
        visible_text = visible_text[:self.dim[1]-1-len(self.prefix)]
        self.win.addstr(0, 0, self.background*(self.dim[1]-1))
        self.win.addstr(0, 0, self.prefix)
        self.win.addstr(0, len(self.prefix), visible_text)
        self.win.noutrefresh()

        if self.completion is not None:
            self.completion.paint(clear)

    def focus(self):
        y = 0
        x = self.cursor - self.offset + len(self.prefix)
        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()
        return y, x

    def scroll(self):
        """Recalculate 'offset' based on window size and cursor position"""
        self.cursor = min(self.cursor, len(self.text))

        if self.margin <= self.cursor - self.offset < self.dim[1] - self.margin - len(self.prefix):
            # no need to change
            return False

        self.offset = max(0, min(len(self.text), self.cursor - self.dim[1] + self.margin + len(self.prefix)))
        return True

    def resize(self, width, compat=None):
        """Change the width of the input field"""
        if compat is not None:
            # some people might call this function with the official (height, width) signature
            # but we're ignoring the height: it's always 1!
            width = compat
        super().resize(1, width)

        if self.completion is not None:
            self.completion.autosize(*self.absolute_cursor)

    def current_word(self):
        """Return the span of the current word (the word at the cursor)

        Returns a tuple (start, end) or None if not on a word.
        """
        start = self.cursor

        if len(self.text[start-1:start+1].strip()) == 0:
            return None

        else:
            if start == len(self.text):
                start -= 1
            elif self.text[start] in string.whitespace:
                start -= 1
            end = start

            # move backwards in the word to find the start
            while start >= 0:
                if self.text[start] in string.whitespace:
                    break
                start -= 1
            start += 1

            # move forward in the word to find the end
            while end < len(self.text) and self.text[end] not in string.whitespace:
                end += 1

        return (start, end)

    def replace_word(self, word, move_to_end=True):
        """Replace the word at the cursor position with ``word``

        If the cursor is on a space, the word is inserted before the cursor.

        Unless ``move_to_end`` is set to ``False``, the cursor will be moved
        to the end of the word at the end of the replacing.
        """
        span = self.current_word()

        if span is None:
            self.text = self.text[:self.cursor] + word + self.text[self.cursor:]
            if move_to_end:
                self.cursor = self.cursor + len(word)

        else:
            start, end = span
            self.text = self.text[:start] + word + self.text[end:]
            if move_to_end:
                self.cursor = start + len(word)

        self.paint()

    @property
    def absolute_cursor(self):
        """Return the (y,x) tuple of the absolute cursor position on screen"""
        return self.pos[0], self.pos[1] + self.cursor - self.offset

    def update_completion(self):
        """Trigger .update on the completion component, if there is any"""
        if self.completion is not None:
            self.completion.update(*self.absolute_cursor)

    def handle_key(self, key):
        must_repaint = False
        current_text = self.text

        if not self.read_only and self.completion is not None and self.completion.handle_key(key):
            logging.debug(f"Completion should have handled {key}")

        elif key in [Key.RIGHT]:
            self.cursor = min(len(self.text), self.cursor+1)
            must_repaint = self.scroll()

        elif key in [Key.LEFT]:
            self.cursor = max(0, self.cursor-1)
            must_repaint = self.scroll()

        elif key in [Key.END, "^E"]:
            self.cursor = len(self.text)
            must_repaint = self.scroll()

        elif key in [Key.HOME, "^A"]:
            self.cursor = 0
            must_repaint = self.scroll()

        elif key in [Key.BACKSPACE, "^H"] and self.cursor > 0 and not self.read_only:
            self.text = self.text[:self.cursor-1] + self.text[self.cursor:]
            self.cursor -= 1
            must_repaint = True

        elif key in ["^U"] and not self.read_only:
            self.text = self.text[self.cursor:]
            self.cursor = 0
            must_repaint = True

        elif key in [Key.DELETE] and len(self.text[self.cursor:]) > 0 and not self.read_only:
            self.text = self.text[:self.cursor] + self.text[self.cursor+1:]
            must_repaint = True

        elif len(str(key)) == 1 and not self.read_only:
            self.text = self.text[:self.cursor] + str(key) + self.text[self.cursor:]
            self.cursor += 1
            must_repaint = True

        if self.completion is not None and self.text != current_text:
            self.update_completion()

        if must_repaint:
            self.paint()
        self.focus()



import curses


class ShellContext:
    """Conveniently allows you to run a subprocess in curses shell mode

    Example usage::

        # in a cursedspace.Application
        with ShellContext(self.screen):
            subprocess.run(...)
        self.paint(True)
    """
    def __init__(self, win, clear=False):
        self.win = win
        self.clear = clear
        self.cursor_state = None

    def __enter__(self, *args, **kwargs):
        if self.clear:
            self.win.clear()
            self.win.refresh()
            curses.doupdate()

        try:
            self.cursor_state = curses.curs_set(1)
        except curses.error:
            pass

        curses.reset_shell_mode()

    def __exit__(self, *args, **kwargs):
        curses.reset_prog_mode()

        self.win.move(0, 0)
        self.win.clear()

        if self.cursor_state is not None:
            curses.curs_set(self.cursor_state)

        self.win.refresh()




#!/usr/bin/env python3
import random
import curses
from cursedspace import Application, Key, Panel, colors, Grid


class Mock(Panel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text = "I'm a Hello Panel"
        self.color = random.randint(1, 8)

    def paint(self, clear=False):
        super().paint(clear)
        y, x, h, w = self.content_area()
        for i in range(h):
            self.win.addstr(y + i, x, self.text[:w], colors.attr(self.color))
        self.win.noutrefresh()


class DemoApplication(Application):
    def __init__(self):
        super().__init__()
        # Instantiate Grid with a reference to the application
        self.grid = Grid(self, rows=4, cols=5)

        # Add some panels, the grid cell declaration can be slices and integers
        self.grid.add_panel(slice(0, 4), slice(0, 2), key='big_mock', panel_class=Mock)
        self.grid.add_panel(0, 2, key='small_mock', panel_class=Mock)
        self.grid.add_panel(1, 2, panel_class=Mock)
        self.grid.add_panel(2, 2, panel_class=Mock)
        self.grid.add_panel(3, 2, panel_class=Mock)
        self.grid.add_panel(0, 3, panel_class=Mock)
        self.grid.add_panel(1, 3, panel_class=Mock)
        self.grid.add_panel(2, 3, panel_class=Mock)
        self.grid.add_panel(3, 3, panel_class=Mock)
        self.grid.add_panel(0, 4, panel_class=Mock)
        self.grid.add_panel(1, 4, panel_class=Mock)
        self.grid.add_panel(2, 4, panel_class=Mock)
        self.grid.add_panel(3, 4, panel_class=Mock)

        # We can easily access all panels in the grid
        for panel in self.grid.panels:
            panel.border = Panel.BORDER_ALL

        # But also use the get-item "key" feature but also index in order of adding to the grid
        self.grid['small_mock'].border = Panel.BORDER_TOP
        self.grid[2].border = Panel.BORDER_BOTTOM
        self.grid[3].border = Panel.BORDER_LEFT
        self.grid[4].border = Panel.BORDER_RIGHT

        # Lets showcase some border combinations
        self.grid[5].border = Panel.BORDER_TOP + Panel.BORDER_LEFT
        self.grid[6].border = Panel.BORDER_TOP + Panel.BORDER_RIGHT
        self.grid[7].border = Panel.BORDER_BOTTOM + Panel.BORDER_LEFT
        self.grid[8].border = Panel.BORDER_BOTTOM + Panel.BORDER_RIGHT
        self.grid[9].border = Panel.BORDER_TOP + Panel.BORDER_BOTTOM
        self.grid[10].border = Panel.BORDER_LEFT + Panel.BORDER_RIGHT
        self.grid[11].border = Panel.BORDER_BOTTOM + Panel.BORDER_LEFT + Panel.BORDER_RIGHT
        self.grid[12].border = Panel.BORDER_TOP + Panel.BORDER_BOTTOM + Panel.BORDER_LEFT

    def draw(self):
        self.grid.paint()
        curses.doupdate()

    def main(self):
        self.resize()

        while True:
            self.draw()
            
            key = self.read_key()
            for panel in self.grid.panels:
                panel.color = random.randint(1, 8)

            if key == Key.RESIZE:
                self.resize()
            elif key in [Key.ESCAPE, "q", "^C"]:
                break

    def resize(self):
        height, width = self.size()
        self.grid.resize(height, width)


# run the application
DemoApplication().run()



#!/usr/bin/env python3
import curses

from cursedspace import Application, Key, InputLine
from cursedspace import SingleWordCompletionWrapper, Completion


class DemoApplication(Application):
    def __init__(self):
        super().__init__()
        self.input = None

    def main(self):
        words = [Completion.Suggestion("Spanish Inquisition", "nobody"),
                 "parrot",
                 "albatross"]
        self.input = InputLine(self,
                               min(self.size()[1], 30),
                               (3, 3),
                               text="Press Escape or ^C to exit",
                               background='_',
                               completion=SingleWordCompletionWrapper(words))
        self.input.paint()
        self.input.focus()

        while True:
            curses.doupdate()

            key = self.read_key()

            if key in [Key.ESCAPE, "^C"]:
                break
            self.input.handle_key(key)

    def refresh(self, force=False):
        super().refresh(force)
        if self.input is not None:
            self.input.paint()


DemoApplication().run()



from .panel import Panel
from . import colors


class ProgressBar(Panel):
    """A generic progressbar that can be updated.
    """

    PROGRESS_SYMBOL = "█"
    NO_PROGRESS_SYMBOL = " "

    def __init__(self, *args, width=None, border=Panel.BORDER_NONE, description='', color=None, **kwargs):
        self.border = border
        if width is None:
            size = None
        else:
            size = (self.height, width)
        kwargs['size'] = size
        kwargs['border'] = border
        super().__init__(*args, **kwargs)

        self.color = color
        self.progress = 0
        self.description = description
        self.description_size = len(description)

    @property
    def height(self):
        h = 1
        if self.border & Panel.BORDER_TOP != 0:
            h += 1
        if self.border & Panel.BORDER_BOTTOM != 0:
            h += 1
        if self.border & (Panel.BORDER_LEFT + Panel.BORDER_RIGHT) != 0:
            h = 3
        return h

    @property
    def top(self):
        if self.border & (Panel.BORDER_LEFT + Panel.BORDER_RIGHT + Panel.BORDER_TOP) != 0:
            return 1
        else:
            return 0

    def update(self, progress, description=None):
        if progress < 0 or progress > 100:
            raise ValueError(f'Progress "{progress}" cannot be less then 0 or higher than 100')
        if description is not None:
            self.description = description[:self.description_size]

        self.progress = progress
        self.refresh()

    def resize(self, h, w):
        super().resize(self.height, w)

    def paint(self, clear=False):
        super().paint(clear=clear)
        _, x, h, w = self.content_area()
        bar_w = w - self.description_size - 2
        y = self.top

        painted = round(bar_w*self.progress/100)
        if painted > bar_w:
            painted = bar_w
        not_painted = bar_w - painted

        bar = self.PROGRESS_SYMBOL*painted + self.NO_PROGRESS_SYMBOL*not_painted

        self.win.addstr(y, x, self.description)
        self.win.addstr(y, x + self.description_size, ' ')
        if self.color is None:
            self.win.addstr(y, x + self.description_size + 1, bar)
        else:
            self.win.addstr(y, x + self.description_size + 1, bar, colors.attr(self.color))

        self.win.noutrefresh()



import setuptools
import pathlib


try:
    import docutils.core
    from docutils.writers import manpage
except ImportError:
    docutils = None
    manpage = None


from cursedspace import version


with open('README.md', encoding='utf-8') as fd:
    long_description = fd.read()


with open('LICENSE', encoding='utf-8') as fd:
    licensetext = fd.read()


def compile_documentation():
    htmlfiles = []

    if docutils is None:
        return htmlfiles

    dst = pathlib.Path('./cursedspace/docs')
    dst.mkdir(exist_ok=True)
    
    pathlib.Path('./man').mkdir(exist_ok=True)

    man_pter = None

    if None not in [docutils, manpage]:
        for fn in pathlib.Path('./doc').iterdir():
            if fn.suffix == '.rst':
                if fn.stem == 'cursedspace':
                    man_pter = str(fn)
                dstfn = str(dst / (fn.stem + '.html'))
                docutils.core.publish_file(source_path=str(fn),
                                           destination_path=dstfn,
                                           writer_name='html')
                htmlfiles.append('docs/' + fn.stem + '.html')

    if man_pter is not None:
        docutils.core.publish_file(source_path=man_pter,
                                   destination_path='man/cursedspace.3',
                                   writer_name='manpage')

    return htmlfiles


def example_files():
    path = pathlib.Path('./examples')
    return list('examples/' + fn.name for fn in path.iterdir() if fn.suffix == '.py')


setuptools.setup(
    name='cursedspace',
    version=version.__version__,
    description="Library for TUI programs on basis of curses",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license_file="LICENSE",
    license_files="LICENSE",
    url="https://vonshednob.cc/cursedspace",
    author="R",
    author_email="contact+cursedspace@vonshednob.cc",
    entry_points={'console_scripts': []},
    packages=['cursedspace'],
    package_data={'cursedspace': compile_documentation()},
    data_files=[('share/man/man3', ['man/cursedspace.3']),
                ('share/applications', []),
                ('share/doc/cursedspace', []),
                ('share/doc/cursedspace/examples', example_files())],
    install_requires=[],
    extras_require={},
    python_requires='>=3.8',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console :: Curses',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3',])



import unittest
import curses

from cursedspace import InputLine


class FakeApp:
    def __init__(self):
        self.screen = FakeWin()


class FakeWin:
    def addstr(self, *args, **kwargs):
        pass

    def move(self, *args, **kwargs):
        pass

    def noutrefresh(self):
        pass

    def clear(self):
        pass

    def mvwin(self, *args, **kwargs):
        pass

    def touchwin(self):
        pass


def newwin(*args, **kwargs):
    return FakeWin()


def nothing():
    pass


curses.newwin = newwin


class TestCurrentWord(unittest.TestCase):
    def setUp(self):
        self.inputline = InputLine(FakeApp(), 20, (0, 0))
        self.inputline.win = FakeWin()
        self.inputline.paint = nothing

    def test_empty_case(self):
        self.inputline.text = ''
        self.inputline.scroll()

        self.assertIsNone(self.inputline.current_word())

    def test_no_word(self):
        self.inputline.text = 'there is  space'
        self.inputline.cursor = 9

        span = self.inputline.current_word()

        self.assertIsNone(span)

    def test_basic_case1(self):
        self.inputline.text = 'this is a test'
        self.inputline.cursor = 5

        span = self.inputline.current_word()

        self.assertEqual(span, (5, 7))

    def test_basic_case2(self):
        self.inputline.text = 'a word'
        self.inputline.cursor = 6

        span = self.inputline.current_word()

        self.assertEqual(span, (2, 6))

    def test_ws_end(self):
        self.inputline.text = 'there be blanks '
        self.inputline.cursor = 16

        span = self.inputline.current_word()

        self.assertIsNone(span)

    def test_ws_start(self):
        self.inputline.text = ' there be blanks'
        self.inputline.cursor = 0

        span = self.inputline.current_word()

        self.assertIsNone(span)

    def test_cursor_after_word(self):
        self.inputline.text = 'this is a test'
        self.inputline.cursor = 7

        span = self.inputline.current_word()

        self.assertEqual(span, (5, 7))


class TestReplaceWord(unittest.TestCase):
    def setUp(self):
        self.inputline = InputLine(FakeApp(), 20, (0, 0))
        self.inputline.win = FakeWin()
        self.inputline.paint = nothing

    def test_empty_case(self):
        self.inputline.text = ''
        self.inputline.scroll()

        self.inputline.replace_word('hello')

        self.assertEqual(self.inputline.text, 'hello')

    def test_insert_case(self):
        self.inputline.text = 'there is  space'
        self.inputline.cursor = 9

        self.inputline.replace_word('enough')

        self.assertEqual(self.inputline.text, 'there is enough space')

    def test_basic_case1(self):
        self.inputline.text = 'this is a test'
        self.inputline.cursor = 5

        self.inputline.replace_word('was')

        self.assertEqual(self.inputline.text, 'this was a test')

    def test_basic_case2(self):
        self.inputline.text = 'this is a test'
        self.inputline.cursor = 6

        self.inputline.replace_word('was')

        self.assertEqual(self.inputline.text, 'this was a test')

    def test_insert1(self):
        self.inputline.text = 'another test'
        self.inputline.cursor = 7
        
        self.inputline.replace_word('funky')

        self.assertEqual(self.inputline.text, 'funky test')

    def test_last_word(self):
        self.inputline.text = 'one more test'
        self.inputline.cursor = 10

        self.inputline.replace_word('forkup')

        self.assertEqual(self.inputline.text, 'one more forkup')

    def test_first_word(self):
        self.inputline.text = 'test this scenario'
        self.inputline.cursor = 2

        self.inputline.replace_word('fudge')

        self.assertEqual(self.inputline.text, 'fudge this scenario')




import curses

from .panel import Panel
from .key import Key


class ScrollPanel(Panel):
    """A generic panel that allows for scrolling content (e.g. a list)

    The content to scroll is per line in self.items.

    After updating the content of self.items you may want to reset
    self.offset and self.cursor to point to a valid element in the list of items.
    Or just set self.cursor and self.scroll to determine the self.offset
    automatically.

    ScrollPanel.SCROLL_MARGIN defines when the list should start scrolling.
    ScrollPanel.SCROLL_PAGING defines whether or not the scrolling should be paging
      instead of line by line.
    ScrollPanel.SCROLL_NEXT and ScrollPanel.SCROLL_PREVIOUS are lists of valid Keys
      that cause the selection of the next/previous item respectively. These can
      trigger scrolling.

    To control how an item is displayed, override the do_paint_item function.
    """
    SCROLL_MARGIN = 5
    SCROLL_PAGING = False
    SCROLL_NEXT = [Key.DOWN]
    SCROLL_PREVIOUS = [Key.UP]
    SCROLL_NEXT_PAGE = [Key.PGDN]
    SCROLL_PREVIOUS_PAGE = [Key.PGUP]
    SCROLL_TO_START = [Key.HOME]
    SCROLL_TO_END = [Key.END]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = []
        self.cursor = 0
        self.offset = 0
        self.list_height = 1
        self.wrapping = False

    @property
    def selected_item(self):
        """The currently selected item"""
        if 0 <= self.cursor < len(self.items):
            return self.items[self.cursor]
        return None

    def do_paint_item(self, y, x, maxwidth, is_selected, item):
        """Here is where you define how to paint that item on the window

        y, x: position where to draw the item
        maxwidth: available space on that line
        is_selected: whether or not that item is the currently selected one
        item: the item to paint
        """
        try:
            self.win.addstr(y, x, str(item)[:maxwidth])
        except curses.error:
            pass

    def paint(self, clear=False):
        super().paint(clear)

        y, x, _, w = self.content_area()

        for itemidx in range(self.offset, min(self.offset + self.list_height, len(self.items))):
            self.paint_item(itemidx)
            y += 1

        # Fill the rest with blanks
        for y in range(y, self.list_height+1):
            try:
                self.win.addstr(y, x, " "*w)
            except curses.error:
                pass

        self.win.noutrefresh()

    def jump_to(self, item):
        """Select this item by instance or by index

        Puts the cursor on this item and returns True (you have to call
        paint yourself; most likely with clear=True).
        Returns False if the item is not in the list or the index was out
        of range.
        """
        if isinstance(item, int):
            if item < 0 or item >= len(self.items):
                return False
        else:
            if item not in self.items:
                return False
            item = self.items.index(item)

        self.cursor = item
        self.scroll()
        return True

    def focus(self):
        y, x, h, _ = self.content_area()
        y = max(y, min(h, y + self.cursor - self.offset))

        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()
        return y, x

    def resize(self, *args):
        super().resize(*args)
        self.calc_list_height()

    def calc_list_height(self):
        _, _, self.list_height, _ = self.content_area()

    def paint_item(self, itemidx):
        if not isinstance(itemidx, int):
            itemidx = self.items.index(itemidx)

        y, x, h, w = self.content_area()
        itemy = y + itemidx - self.offset

        if itemy < y or itemy > h:
            return

        self.do_paint_item(itemy, x, w, itemidx == self.cursor, self.items[itemidx])

    def scroll(self):
        self.cursor = max(0, min(self.cursor, len(self.items)-1))
        scroll_margin = min(self.list_height // 2, self.SCROLL_MARGIN)

        if self.SCROLL_PAGING:
            if self.cursor - self.offset >= self.list_height - scroll_margin and self.list_height < len(self.items):
                self.offset = max(0, self.cursor - scroll_margin)
            elif self.cursor - self.offset < scroll_margin and self.offset > 0:
                self.offset = max(0, self.cursor - self.list_height + scroll_margin + 1)
            else:
                return False
        else:
            if self.cursor - self.offset < scroll_margin:
                self.offset = max([0,
                                   self.cursor - self.list_height,
                                   self.cursor - scroll_margin])
            if self.cursor - self.offset >= self.list_height - scroll_margin:
                self.offset = max(0, self.cursor + min(scroll_margin, len(self.items)-self.cursor) - self.list_height)
            else:
                return False

        return True

    def handle_key(self, key):
        """Handle user input key

        Returns True if handled, False otherwise
        """
        handled, must_repaint, must_clear = self.handle_scrolling_keys(key)

        if handled and must_repaint:
            self.paint(must_clear)

        return handled

    def select_next(self):
        """Move the cursor to the next item and update the representation

        Returns True if repaint required, otherwise False"""
        if self.wrapping:
            index = (self.cursor+1) % len(self.items)
        else:
            index = min(self.cursor+1, len(self.items)-1)
        return self._select_by_index(index)

    def select_previous(self):
        """Move the cursor to the previous item and update the representation

        Returns True if repaint required, otherwise False"""
        if self.wrapping:
            index = (self.cursor-1) % len(self.items)
        else:
            index = max(self.cursor-1, 0)
        return self._select_by_index(index)

    def _select_by_index(self, index):
        """Move the cursor to item with index

        Return True if repaint is required, False otherwise

        If False is returned the list may or may not have updated
        and repainted the old and the new element.
        """
        must_repaint = False
        if index != self.cursor:
            old_cursor = self.cursor
            self.cursor = index
            must_repaint = self.scroll()
            if not must_repaint:
                self.paint(old_cursor)
                self.paint(self.cursor)
        return must_repaint

    def handle_scrolling_keys(self, key):
        """Can be called to handle scrolling keys

        If a scrolling key is found, it will likely update self.cursor.

        May call self.scroll (which updates self.offset).
        May call self.paint_item (and self.do_paint_item subsequently).

        Returns a tuple (handled, must_repaint, must_clear) meaning this:
            * handled: the key was a scrolling key
            * must_repaint: you should call self.paint
            * must_clear: you should pass clear=True to self.paint
        """
        must_repaint = False
        must_clear = False
        cursor = self.cursor
        handled = False

        if key in self.SCROLL_PREVIOUS:
            if cursor <= 0:
                if self.wrapping:
                    cursor = len(self.items) - 1
            else:
                cursor -= 1
            handled = True

        elif key in self.SCROLL_NEXT:
            if cursor == len(self.items)-1:
                if self.wrapping:
                    cursor = 0
            else:
                cursor += 1
            handled = True

        elif key in self.SCROLL_PREVIOUS_PAGE and cursor > 0:
            cursor = max(0, cursor - self.list_height)
            handled = True

        elif key in self.SCROLL_NEXT_PAGE and cursor < len(self.items)-1:
            cursor = min(len(self.items)-1, cursor + self.list_height)
            handled = True

        elif key in self.SCROLL_TO_END:
            cursor = len(self.items)-1
            handled = True

        elif key in self.SCROLL_TO_START:
            cursor = 0
            handled = True

        if handled:
            must_repaint = self._select_by_index(cursor)
            must_clear = True

        return handled, must_repaint, must_clear



import curses


class Panel:
    """A generic panel with or without border"""
    BORDER_NONE = 0
    BORDER_TOP = 1
    BORDER_BOTTOM = 2
    BORDER_LEFT = 4
    BORDER_RIGHT = 8
    BORDER_ALL = 15
    BORDER_STYLE = '┌┐└┘─│'

    def __init__(self, app, size=None, pos=None, border=None):
        """app is the cursedspace.Application this panel belongs to
           size is a tuple (height, width)
           pos the position on the application's window (y, x)"""
        self.app = app
        self.dim = (0, 0)
        self.pos = (0, 0)
        self.win = None
        self.border = Panel.BORDER_NONE if border is None else border

        if size is not None:
            self.resize(*size)
            if pos is not None:
                self.move(*pos)

    def destroy(self):
        """Should be called explicitely when closing a panel to restore
        the window(s) "behind" this panel."""
        if self.win is not None:
            del self.win
            self.win = None
            # self.app.refresh(True)

    def __delete__(self):
        self.destroy()

    def handle_key(self, key):
        """Panel specific call-back when a key has been entered by the user

        key is of type cursedspace.Key"""
        pass

    def resize(self, h, w):
        """Resize the panel"""
        self.dim = (h, w)
        if self.win is None:
            self.win = curses.newwin(*self.dim, *self.pos)
        else:
            self.win.resize(*self.dim)
        self.win.clear()
        self.win.noutrefresh()

    def refresh(self, force=False):
        """Refresh the panel

        force will touch the panel's content prior to refreshing"""
        if self.win is None:
            return
        if force:
            self.win.touchwin()
        self.win.noutrefresh()

    def content_area(self):
        """The area available for content, i.e. inside the borders

        Returns a tuple (y, x, height, width)
        """
        y = 0
        x = 0
        h, w = self.dim

        if self.border & Panel.BORDER_TOP != 0:
            y += 1
            h -= 1
        if self.border & Panel.BORDER_LEFT != 0:
            x += 1
            w -= 1
        if self.border & Panel.BORDER_RIGHT != 0:
            w -= 1
        if self.border & Panel.BORDER_BOTTOM != 0:
            h -= 1

        return y, x, h, w

    def focus(self):
        """Move the cursor into this panel"""
        y, x, _, _ = self.content_area()
        
        try:
            self.win.move(y, x)
        except curses.error:
            pass
        self.win.noutrefresh()

        return y, x

    def paint(self, clear=False):
        """Draws the panel into the curses buffer

        Will draw a border based on what self.border is set to.
        If clear is True, it will erase any content of the window first."""
        top_left = 0
        bottom_left = 0
        top_width = self.dim[1]
        bottom_width = self.dim[1]
        left_height = self.dim[0]
        right_height = self.dim[0]
        left_top = 0
        right_top = 0
        corner_top_left = Panel.BORDER_TOP + Panel.BORDER_LEFT
        corner_top_right = Panel.BORDER_TOP + Panel.BORDER_RIGHT
        corner_bottom_left = Panel.BORDER_BOTTOM + Panel.BORDER_LEFT
        corner_bottom_right = Panel.BORDER_BOTTOM + Panel.BORDER_RIGHT

        if clear:
            self.win.erase()

        if self.border & corner_top_left != 0:
            corner = self.BORDER_STYLE[5]
            if self.border & corner_top_left == corner_top_left:
                corner = self.BORDER_STYLE[0]
            elif self.border & corner_top_left == Panel.BORDER_TOP:
                corner = self.BORDER_STYLE[4]
            self.win.addstr(0, 0, corner)
            left_height -= 1
            left_top += 1
            top_width -= 1
            top_left += 1

        if self.border & corner_top_right != 0:
            corner = self.BORDER_STYLE[5]
            if self.border & corner_top_right == corner_top_right:
                corner = self.BORDER_STYLE[1]
            elif self.border & corner_top_right == Panel.BORDER_TOP:
                corner = self.BORDER_STYLE[4]
            self.win.addstr(0, self.dim[1]-1, corner)
            right_height -= 1
            right_top += 1
            top_width -= 1

        if self.border & corner_bottom_left != 0:
            corner = self.BORDER_STYLE[5]
            if self.border & corner_bottom_left == corner_bottom_left:
                corner = self.BORDER_STYLE[2]
            elif self.border & corner_bottom_left == Panel.BORDER_BOTTOM:
                corner = self.BORDER_STYLE[4]
            self.win.addstr(self.dim[0]-1, 0, corner)
            left_height -= 1
            bottom_width -= 1
            bottom_left += 1

        if self.border & corner_bottom_right != 0:
            # curses raises an exception when drawing in the lowest, most right
            # cell of the window
            corner = self.BORDER_STYLE[5]
            if self.border & corner_bottom_right == corner_bottom_right:
                corner = self.BORDER_STYLE[3]
            elif self.border & corner_bottom_right == Panel.BORDER_BOTTOM:
                corner = self.BORDER_STYLE[4]
            try:
                self.win.addstr(self.dim[0]-1, self.dim[1]-1, corner)
            except curses.error:
                pass
            right_height -= 1
            bottom_width -= 1

        if self.border & Panel.BORDER_TOP != 0:
            self.win.addstr(0, top_left, self.BORDER_STYLE[4]*top_width)

        if self.border & Panel.BORDER_BOTTOM != 0:
            self.win.addstr(self.dim[0]-1, bottom_left, self.BORDER_STYLE[4]*bottom_width)

        if self.border & Panel.BORDER_LEFT != 0:
            for y in range(left_height):
                self.win.addstr(left_top+y, 0, self.BORDER_STYLE[5])

        if self.border & Panel.BORDER_RIGHT != 0:
            for y in range(right_height):
                self.win.addstr(right_top+y, self.dim[1]-1, self.BORDER_STYLE[5])

        self.win.noutrefresh()

    def move(self, y, x):
        """Move the panel"""
        if self.win is None:
            return
        self.pos = (y, x)
        self.win.mvwin(y, x)
        self.app.screen.touchwin()
        self.app.screen.noutrefresh()
